local jutils = require("utility.jutils")
local config = require("config")
local tiles = require("tiles")
local playerinventory = require("inventory.playerinventory")
local grid = require("utility.grid")
local input = require("utility.input")
local items = require("items")
local rendering = require("utility.rendering")

local physicalentity = require("entities.physicalentity")

local player = physicalentity:subclass("Player")

function player:init()
	physicalentity.init(self)

	self.texture = love.graphics.newImage("assets/entities/player.png")
	self.boundingbox = jutils.point.new(6, 12)
	self.textureorigin = jutils.point.new(8, 12)
	self.fallthrough = false
	self.mass = 1.25
	self.direction = -1 -- -1=left 0=forward 1=right
	
	self.animationframes = {
		[1] = love.graphics.newQuad(0, 0, 16, 24, 64, 24),-- standing still
		[2] = love.graphics.newQuad(16, 0, 16, 24, 64, 24),-- walking1,
		[3] = love.graphics.newQuad(32, 0, 16, 24, 64, 24),--walking2
		[4] = love.graphics.newQuad(48, 0, 16, 24, 64, 24),--walking2
	}

	self.freecamspeed = 4
	self.inventory = playerinventory:new()
	self.inventory.player = self
	self.inventoryOpen = false
	self.walkspeed = 70
	self.acceleration = 900
	self.climbcooldown = 0
	self.showMouseTileDistance = -1
	self.hotbarslot = 1
	self.currentSelectedSlot = 0
	self.itemEquipped = nil
	self.lastItemEquipped = nil
	self.equipDebounce = false
	self.fast = false
	self.statuseffects = {}
	self.mouse = {
		leftDown = false, rightDown = false, 
		leftBounce = false, rightBounce = false,
		position = jutils.point.new(0, 0)
	}

	self.lclicked = false
	self.rclicked = false
	self.lusecooldown = 0
	self.rusecooldown = 0
	self.effects = {}
	self.escapeDebounce = false

	self.touchingrope = false
	self.grabrope = false
	self.onrope = false

	self.animation = {
		swingtimer = 0,
		swinging = false,
		item = nil,
		walking = false,
		walkstep = 1,
	}

	self.inventory:addItem(items.HANDMADE_PICKAXE.id, 1)
	self.inventory:addItem(items.TNT_TILE.id, 99)
	self.inventory:addItem(items.ACTIVATOR_TILE.id, 99)
	self.inventory:addItem(items.ROPE_TILE.id, 999)
	self.inventory:addItem(items.HEALING_POTION.id, 30)

	-- ghetto retard workaround for emptying input callbacks when player is destroyed
	self.inputcallbacks = {
		mousepressed = input.mousepressed:connect(function(...)
			self:onMousePressed(...)
		end),
		mousereleased = input.mousereleased:connect(function(...)
			self:onMouseReleased(...)
		end),
		wheelmoved = input.wheelmoved:connect(function(...)
			self:onWheelMoved(...)
		end),
		keypressed = input.keypressed:connect(function(...)
			self:onKeyPressed(...)
		end),
	}
	
	self:scroll(0)
end

function player:dies()
	
	for idx, id in pairs(self.inputcallbacks) do
		input[idx]:disconnect(id)
	end
	self.inputcallbacks = nil
	physicalentity.dies(self)
end

function player:onMousePressed(x, y, button, istouch, presses)
	local result = self.inventory:clicked(x, y, button, istouch, presses)
	if result == false then
		if button == 1 then self.mouse.leftDown = true end
		if button == 2 then self.mouse.rightDown = true end
	end

	local mousex, mousey = grid.pixelToTileXY(input.getTransformedMouse())

	local tile = self.world:getTile(mousex, mousey)

	local tiledata = tiles:getByID(tile)

	if tiledata.playerInteract then
		tiledata.playerInteract(self, mousex, mousey, button)
	end
end

function player:onMouseReleased(x, y, button, istouch, presses)
	if button == 1 then self.mouse.leftDown = false end
	if button == 2 then self.mouse.rightDown = false end
	--self:scroll(0)
end

function player:onKeyPressed(key)
	if key == "escape" then
		self.inventoryOpen = not self.inventoryOpen

		if self.inventoryOpen then
			self.inventory:fullOpen()
		else
			self.inventory:fullClose()
		end
	end

	-- item dropping, TODO: make it's own method
	if key == "q" then
		local itemdata = self.inventory:getHeldItem()
		if itemdata then
			local stack = self.itemEquippedStack
			if stack then
				local itemstack = self.world:addEntity("itemstack", stack[1], stack[2])

				itemstack.nextposition = self.nextposition
				itemstack.velocity.y = -120
				itemstack.velocity.x = self.direction*200
				itemstack.playermagnet = -1

				stack[1] = 0
				stack[2] = 0
			end
		end
	end
end

function player:canLeftUse()
	if self.itemEquipped == nil then return false end
	if self.itemEquipped.leftuse == nil then return false end

	if self.lusecooldown <= 0 then
		if self.itemEquipped.repeating == true or self.mouse.leftBounce == false then
			return true
		end
	end
	return false
end

function player:canRightUse()
	if self.itemEquipped == nil then return false end
	if self.itemEquipped.rightuse == nil then return false end

	if self.rusecooldown <= 0 then
		if self.itemEquipped.repeating == true or self.mouse.rightBounce == false then
			return true
		end
	end
	return false
end

function player:leftuse()
	local item = self.itemEquipped
	self.lusecooldown = item.speed
	local success = item:leftuse(self)
	if success then
		self.animation.swingtimer = item.speed
		self.animation.swinging = true
		self.animation.item = self.itemEquipped
	end
end

function player:rightuse()
	local item = self.itemEquipped
	self.rusecooldown = item.speed
	item:rightuse(self)

	self.animation.swingtimer = item.speed
	self.animation.swinging = true
	self.animation.item = self.itemEquipped
end

function player:onWheelMoved(x, y)
	self:scroll(y)
end

function player:getMousePosition()
	return self.mouse.position
end

function player:scroll(change)
	self.hotbarslot = self.hotbarslot - change

	if self.hotbarslot < 1 then self.hotbarslot = 10 end
	if self.hotbarslot > 10 then self.hotbarslot = 1 end

	self.inventory.hotbarslot = self.hotbarslot
end

function player:collisionCallback(tileid, tilepos, separation, normal)

	if tiles:tileHasTag(tileid, "platformtile") then
		if normal.x == 0 and normal.y == -1 and self.fallthrough == false then

			local bottom = self.nextposition.y + self.boundingbox.y
			local realbottom = self.position.y + self.boundingbox.y

			local blockbottom = tilepos.y*config.TILE_SIZE+(config.TILE_SIZE/2)

			if realbottom <= blockbottom then
				self.nextposition = self.nextposition + separation
				self.falling = false
				self.velocity.y = 0
			end
		end
	end	

	if tiles:getByID(tileid).solid == true then

		if normal.y == 0 and normal.x ~= 0 then
			if self.velocity.y <= 0 and math.abs(self.velocity.x) > 10 then
				
				local tryY = tilepos.y*config.TILE_SIZE
				if tryY > self.nextposition.y and self.climbcooldown <= 0 then
					self.velocity.x = self.velocity.x * 0.6
					self.nextposition.y = (tryY-self.boundingbox.y)
					self.climbcooldown = 0.05
				end
			end
		end
	end
end

function player:addStatusEffect(effectid, duration)
	table.insert(self.statuseffects, {time = duration, id = effectid, handled = false})
end


function player:inventoryUpdate(dt)
	local itemid = self.inventory:getHeldItem()
	if itemid and itemid[1] > 0 then
		self.itemEquipped = items:getByID(itemid[1])
		self.itemEquippedStack = itemid
	else
		self.itemEquipped = nil
		self.itemEquippedStack = nil
	end

	if self.itemEquipped ~= self.lastItemEquipped then
		if self.lastItemEquipped and self.lastItemEquipped.unequip then
			self.lastItemEquipped:unequip(self)
		end
		if self.itemEquipped and self.itemEquipped.equip then
			self.itemEquipped:equip(self)
		end
		self.lastItemEquipped = self.itemEquipped
	end
end

function player:animation_update(dt)
	self.animation.swingtimer = self.animation.swingtimer - dt
	if self.animation.swingtimer <= 0 then
		self.animation.swinging = false
	end

	if self.animation.walking then
		self.animation.walkstep = (self.animation.walkstep + (dt* (math.abs(self.velocity.x)/10) ))
		if self.animation.walkstep >= 3 then
			self.animation.walkstep = 1
		end
	end
end

local effectlist = require("statuseffects")

function player:statuseffect_update(dt)
	for idx, effect in pairs(self.statuseffects) do
		local effectdata = effectlist[effect.id]

		if effect.handled == false then
			effectdata.comeup(self)
			effect.handled = true
		end
		effect.time = effect.time - dt
		effectdata.tick(self, dt)
		
		if effect.time <= 0 then
			effectdata.comedown(self)
			effect = nil
			self.statuseffects[idx] = nil
		end
	end
end

function player:fell(distance)
	distance = math.floor(distance/config.TILE_SIZE)
	

	if distance > 12 then
		print("playerfell", distance)
		self.health = self.health - math.sqrt(distance)*4
	end
end

function player:update(dt)
	self:inventoryUpdate(dt)
	self:animation_update(dt)
	self:statuseffect_update(dt)
	physicalentity.update(self, dt)
	

	-- item usage
	self.lusecooldown = self.lusecooldown - dt
	self.rusecooldown = self.rusecooldown - dt

	if self.mouse.leftDown then
		if self:canLeftUse() then
			self.mouse.leftBounce = true
			self:leftuse()
		else
			self.mouse.leftBounce = false
		end
	end

	if self.mouse.rightDown then
		if self:canRightUse() then
			self.mouse.rightBounce = true
			self:rightuse()
		else
			self.mouse.rightBounce = false
		end
	end

	self.climbcooldown = self.climbcooldown - dt
	self.velocity.x = jutils.math.clamp(-self.walkspeed, self.velocity.x, self.walkspeed)
	self.mouse.position = jutils.point.new(input.getTransformedMouse())

	local mm = (self.fast==true) and 3 or 1

	-- rope climbing
	if self.touchingrope == true then
		if self.onrope == false then
			if self.grabrope == true then
				self.onrope = true
			end
		end
	else
		self.onrope = false
	end

	self.applyGravity = not self.onrope
	self.animation.walking = false
	if self.frozen then
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_UP) then self.nextposition.y = self.nextposition.y - self.freecamspeed*mm end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_DOWN) then self.nextposition.y = self.nextposition.y + self.freecamspeed*mm end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_LEFT) then self.nextposition.x = self.nextposition.x - self.freecamspeed*mm end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_RIGHT) then self.nextposition.x = self.nextposition.x + self.freecamspeed*mm end
	else
		local canJump = true

		if self.falling == true then canJump = false end
		if self.onrope == true then canJump = true end

		-- jumping
		if love.keyboard.isDown(config.keybinds.PLAYER_JUMP) and canJump then 
			self.velocity.y = -200
			self.falling = true
			self.onrope = false
		end

		-- rope and platform controls
		self.grabrope = love.keyboard.isDown(config.keybinds.PLAYER_MOVE_UP)
		self.fallthrough = love.keyboard.isDown(config.keybinds.PLAYER_MOVE_DOWN)

		if self.onrope == true then
			if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_UP) then self.velocity.y = self.velocity.y - (self.acceleration*mm) * dt end
			if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_DOWN) then self.velocity.y = self.velocity.y + (self.acceleration*mm) * dt end

			--if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_LEFT) then self.nextposition.x = self.nextposition.x - 4 end
			--if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_RIGHT) then self.nextposition.x = self.nextposition.x + 4 end

			local yforce = (self.velocity.y * config.physics.FRICTION)
			self.velocity.y = self.velocity.y - yforce

			self.nextposition.x = math.floor(self.nextposition.x /config.TILE_SIZE)*config.TILE_SIZE + (config.TILE_SIZE/2)
		else
			-- left<->right movement
			if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_LEFT) then
				self.direction = -1
				self.animation.walking = true
				self.velocity.x = self.velocity.x - (self.acceleration*mm) * dt
			end
			if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_RIGHT) then
				self.direction = 1
				self.animation.walking = true
				self.velocity.x = self.velocity.x + (self.acceleration*mm) * dt
			end
		end

	end
	self.touchingrope = false
	self.inventory:update(dt)
end

local function magnitude(x1, y1, x2, y2)
	return math.sqrt( (x2 - x1)^2 + (y2-y1)^2 )
end

local oscillator = 0

function player:draw()

	oscillator = oscillator + (1/20)

	local mousex, mousey = grid.pixelToTileXY(input.getTransformedMouse())

	local tile = self.world:getTile(mousex, mousey)
	if tile > 0 then
		local tiledata = tiles:getByID(tile)

		if tiledata.playerInteract then
			love.graphics.setColor(1, 1, 0, (math.sin(oscillator)+1)/2)

			-- TODO: make glow cover the whole object, instead of one tile
			love.graphics.rectangle("line", mousex*config.TILE_SIZE, mousey*config.TILE_SIZE, config.TILE_SIZE, config.TILE_SIZE)
		end
	end

	if math.ceil(self.health) < self.maxhealth then
		rendering:setFont("pixelade_tiny")
		love.graphics.setColor(0.8, 0, 0)
		love.graphics.printf(math.ceil(self.health).."/"..math.ceil(self.maxhealth).."HP", self.position.x-20, self.position.y-self.boundingbox.y-10, 40, "center")
		rendering:setFont("default")
	end
	
	local actionframe = 1

	if self.falltime > 1/20 then
		actionframe = 4
	elseif self.animation.walking then
		actionframe = actionframe + math.floor(self.animation.walkstep)
	end

	love.graphics.setColor(1,1,1)
	love.graphics.draw(
		self.texture,
		self.animationframes[actionframe],
		self.position.x, self.position.y, self.rotation, self.scale.x* (-self.direction), self.scale.y,
		self.textureorigin.x, self.textureorigin.y
	)

	if self.animation.swinging then
		
		local data = self.animation.item
		local item = data.id
		if item>0 then
			local percent = 1-(self.animation.swingtimer/data.speed)
			local startang = data.swing.start
			local dist = data.swing.distance
			local currentDist = (percent*dist)
			local angleFixed = (startang + currentDist) * self.direction
			local adjusted = math.rad(data.defaultRotation+angleFixed)
			
			-- radius, no fucking clue why I have to flip this as well
			local r = 8 * self.direction
			-- shift item by dx and dy for swivel effect
			local dx, dy = math.cos(math.rad(angleFixed))*r, math.sin(math.rad(angleFixed))*r

			rendering:drawItem(item, self.position.x+dx, self.position.y+dy, data.inWorldScale, adjusted, data.playerHoldPosition.x, data.playerHoldPosition.y)
		end
	end
	
	local mouse = self:getMousePosition()
	local tmx, tmy = grid.pixelToTileXY(mouse.x, mouse.y)
	local px, py = grid.pixelToTileXY(self.position.x, self.position.y)
	if magnitude(tmx, tmy, px, py) <= self.showMouseTileDistance then
		love.graphics.setColor(1, 1, 1, 0.5)
		love.graphics.setLineWidth(1)
		love.graphics.rectangle("line", tmx*config.TILE_SIZE, tmy*config.TILE_SIZE, config.TILE_SIZE, config.TILE_SIZE)
	end

	self.inventory:draw()
end

return player