local grid = require("utility.grid")
local input = require("utility.input")
local tiles = require("tiles")
local object = require("utility.object")
local jutils = require("utility.jutils")
local collision = require("utility.collision")

local itemcollector = {}
local items = {}
local idinc = 1

love.graphics.setDefaultFilter("nearest", "nearest")

do
	function itemcollector:getByID(itemid)
		for name, data in pairs(items) do
			if data.id == itemid then
				return data
			end
		end
		error("No tile with ID".. itemid.. "exists!")
	end

	function itemcollector:getNumberOfItems()
		return idinc
	end

	local mt = {
		__index = function(t, k)
			if items[k] then return items[k] end
		end
	}
	setmetatable(itemcollector, mt)
end
---------------------------------------------------

local baseitem = object:subclass("Item")

function baseitem:filldefaults()
	self.name = "GENERIC_ITEM"
	self.displayname = "generic.Item"
	self.stack = 1
	self.repeating = true
	self.speed = 1
	self.rarity = 1
	self.id = idinc
	self.texture = love.graphics.newImage("assets/default.png")
	self.color = {1,1,1}
	self.animation = nil
	self.animationspeed = 0
	self.defaultRotation = 0
	self.inWorldScale = 2
	self.playerHoldPosition = jutils.point.new(4, 4)
	self.swing = {
		start = -45,
		distance = 90,
	}

	self.equip = function(self, player) end
	self.unequip = function(self, player) end
	self.leftuse = function(self, player) end
	self.rightuse = function(self, player) end
end

function baseitem:init(name, data)
	self:filldefaults()
	self.name = name
	self.displayname = name

	local latestID = idinc
	if data.id~=nil then
		latestID = data.id
	end

	--print("regitem", name, latestID)

	idinc = latestID + 1

	-- override defaults
	for label, value in pairs(data) do
		if label == "texture" and type(value) == "string" then
			self["texture"] = love.graphics.newImage("assets/items/"..value)
		else
			self[label] = value
		end
	end

	items[name] = self
end

-------------------------------------

local function getMouseTile(world)
	local dx, dy = world:getMousePoint()
	local mx, my = grid.pixelToTileXY(dx, dy)
	return mx, my
end

local function getPlayerTile(playerentity)
	local pos = playerentity.position

	local tx, ty = grid.pixelToTileXY(pos.x, pos.y)
	return tx, ty
end

local function magnitude(x1, y1, x2, y2)
	return math.sqrt( (x2 - x1)^2 + (y2-y1)^2 )
end

local function pickaxeUse(pickaxe, player)
	local mx, my = input.getTransformedMouse()

	local tx, ty = grid.pixelToTileXY(mx, my)
	local px, py = getPlayerTile(player)

	if magnitude(px, py, tx, ty) < pickaxe.usedistance then

		local tileid = player.world:getTile(tx, ty)
		if tileid > 0 then
			
			player.world:damageTile(tx, ty, pickaxe.strength)
			return true
		end
	end
end

baseitem:new("MANLET_POTION", {
	displayname = "MANLET POTION",
	texture = "potion.png",
	color = {0.5, 1, 0.5},
	rarity = 5,
	speed = 1,
	inWorldScale = 1,
	tooltip = [[
Temporary shrinkage in a bottle.
Duration: 60 seconds		
	]],
	swing = {
		start = 0,
		distance = -45,
	},
	playerHoldPosition = jutils.point.new(4, 4),
	defaultRotation = math.rad(0),
	stack = 30,
	repeating = false,
	leftuse = function(self, player)
		local stack = player.itemEquippedStack
		stack[2] = stack[2]-1
		player:addStatusEffect("MANLET", 60)
		return true
	end,
})

baseitem:new("SPEED_POTION", {
	displayname = "QUICKFOOT POTION",
	texture = "potion.png",
	color = {0.5, 0.5, 1},
	rarity = 4,
	speed = 1,
	inWorldScale = 1,
	tooltip = [[
"Velocity increase guaranteed, or your money back!"
Duration: 60 seconds		
	]],
	swing = {
		start = 0,
		distance = -45,
	},
	playerHoldPosition = jutils.point.new(4, 4),
	defaultRotation = math.rad(0),
	stack = 30,
	repeating = false,
	leftuse = function(self, player)
		local stack = player.itemEquippedStack
		stack[2] = stack[2]-1
		player:addStatusEffect("ACCELLERATION", 60)
		return true
	end,
})

baseitem:new("LOWMASS_POTION", {
	displayname = "WEIGHT LOSS SERUM",
	texture = "potion.png",
	color = {1, 0.25, 0.25},
	rarity = 4,
	speed = 1,
	inWorldScale = 1,
	tooltip = [[
"DANGER: Do not mix with caffeine!"
Duration: 60 seconds		
	]],
	swing = {
		start = 0,
		distance = -45,
	},
	playerHoldPosition = jutils.point.new(4, 4),
	defaultRotation = math.rad(0),
	stack = 30,
	repeating = false,
	leftuse = function(self, player)
		local stack = player.itemEquippedStack
		stack[2] = stack[2]-1
		player:addStatusEffect("LOWMASS", 60)
		return true
	end,
})

baseitem:new("ENERGY_DRINK", {
	displayname = "YEET(tm) ENERGY DRINK",
	texture = "potion.png",
	color = {1, 1, 0.25},
	rarity = 4,
	speed = 1,
	inWorldScale = 1,
	tooltip = [[
"Not safe for human consumption"	
	]],
	swing = {
		start = 0,
		distance = -45,
	},
	playerHoldPosition = jutils.point.new(4, 4),
	defaultRotation = math.rad(0),
	stack = 30,
	repeating = false,
	leftuse = function(self, player)
		local stack = player.itemEquippedStack
		stack[2] = stack[2]-1
		player:addStatusEffect("EXPLOSION", 5)
		return true
	end,
})

baseitem:new("HEALING_POTION", {
	displayname = "GENERIC HEALTH POTION",
	texture = "potion.png",
	color = {1, 0.5, 1},
	rarity = 2,
	speed = 1,
	inWorldScale = 1,
	tooltip = [[
Restores 5 HP/second.
Duration 10 seconds.
	]],
	swing = {
		start = 0,
		distance = -45,
	},
	playerHoldPosition = jutils.point.new(4, 4),
	defaultRotation = math.rad(0),
	stack = 30,
	repeating = false,
	leftuse = function(self, player)
		local stack = player.itemEquippedStack
		stack[2] = stack[2]-1
		player:addStatusEffect("HEAL", 10)
		return true
	end,
})

baseitem:new("WEAPON", {
	displayname = "TEST WEAPON",
	speed = 1/2,
	texture = "sword.png",
	color = {1,0,0},
	tooltip = "working bruh",
	stack = 1,
	inWorldScale = 2,
	swing = {
		start = -90,
		distance = 135
	},
	playerHoldPosition = jutils.point.new(0, 4),
	defaultRotation = 0,
	equip = function(self, player)

	end,
	unequip = function(self, player)

	end,
	leftuse = function(self, player)
		local dir = player.direction

		local hitbox = {}

		-- figure out hitbox overlaps
		local world = player.world

		for _, entity in pairs(world.entities) do
			if entity ~= player then
				local pos = entity.position
				local box = entity.boundingbox

				if collision:aabb(player.position.x+8, player.position.y, 16, 2, pos.x, pos.y, box.x, box.y) then
					print("cum")
					entity.health = entity.health - 5
				end
			end
		end
		return true
	end,
})

local PICKAXE_TOOLTIP = 
[[
{speed} Speed
{usedistance} Range
{strength} Pickaxe Power
]]

baseitem:new("HANDMADE_AXE", {
	displayname = "HANDMADE AXE",
	texture = "pickaxenew.png",
	speed = 1/4,
	tooltip = "Crude yet effective handtool.\n"..PICKAXE_TOOLTIP,
	stack = 1,
	usedistance = 6,
	strength = 1,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

})

baseitem:new("HANDMADE_PICKAXE", {
	displayname = "HANDMADE PICKAXE",
	texture = "pickaxenew.png",
	stack = 1,
	tooltip = "A rock and a stick binded with rope.\n"..PICKAXE_TOOLTIP,
	color = {0.6, 0.6, 0.6},
	speed = 1/4,
	usedistance = 6,
	swing = {
		start = -45,
		distance = 180
	},
	playerHoldPosition = jutils.point.new(0, 8),
	strength = 1,
	defaultRotation = -45,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = pickaxeUse,
})

baseitem:new("IRON_PICKAXE", {
	texture = "pickaxenew.png",
	color = {0.9, 0.75, 0.75},
	speed = 1/6,
	tooltip = PICKAXE_TOOLTIP,
	stack = 1,
	swing = {
		start = math.pi/2,
		distance = 1.5*math.pi
	},
	playerHoldPosition = jutils.point.new(0, 8),
	rarity = 2,
	usedistance = 8,
	strength = 2.5,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

	leftuse = pickaxeUse,
})

baseitem:new("COPPER_PICKAXE", {
	texture = "pickaxenew.png",
	color = {0.9, 0.45, 0.1},
	speed = 1/10,
	tooltip = PICKAXE_TOOLTIP,
	stack = 1,
	swing = {
		start = math.pi/2,
		distance = 1.5*math.pi
	},
	playerHoldPosition = jutils.point.new(0, 8),

	usedistance = 10,
	strength = 1,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = pickaxeUse,
})

baseitem:new("LEAD_PICKAXE", {
	texture = "pickaxenew.png",
	color = {0.35, 0.35, 0.45},
	speed = 1/4,
	tooltip = PICKAXE_TOOLTIP,
	stack = 1,
	swing = {
		start = math.pi/2,
		distance = 1.5*math.pi
	},
	playerHoldPosition = jutils.point.new(0, 8),
	usedistance = 12,
	strength = 4,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = pickaxeUse,
})

local function multiTilePlace(self, player)
		
	local mx, my = input.getTransformedMouse()
	local stack = player.itemEquippedStack
	local tx, ty = grid.pixelToTileXY(mx, my)

	local function isempty(world, x, y)
		if world:getTile(x, y) == 0 then return true end
	end

	local function test(offx, offy)
		for tilename, position in pairs(self.tileset) do
			if not isempty(player.world, tx+offx+position[1], ty+offy+position[2]) then
				return false
			end
		end
		return true
	end

	local doAnyWork = false
	local xrange = (self.tilesize[1])
	local yrange = (self.tilesize[2])
	for x = 0, xrange-1 do
		for y = 0, yrange-1 do

			local realx = math.floor(xrange/2)
			local realy = math.floor(yrange/2)

			if test(x-xrange, y-yrange) then
				doAnyWork = true
				for tilename, position in pairs(self.tileset) do
					print(tilename)
					local tileid = tiles[tilename].id

					if tiles[tilename].validplacement then
						local result = tiles[tilename].validplacement(player.world, (tx+x+position[1]) - xrange, (ty+y+position[2]) - yrange)

						if result == false then
							return 
						end
					end

					player.world:setTile((tx+x+position[1]) - xrange, (ty+y+position[2]) - yrange, tileid)
				end

				stack[2] = stack[2] - 1
				return
			end
		end
	end
end

baseitem:new("CAMPFIRE", {
	stack = 99,
	usedistance = 8,
	speed = 1/10,
	tilesize = {2, 1},
	tileset = {
		CAMPFIRE_1_1 = {1,1},
		CAMPFIRE_2_1 = {2,1},
	},
	leftuse = multiTilePlace
})

baseitem:new("FURNACE", {
	stack = 99,
	usedistance = 8,
	speed = 1/10,
	tilesize = {2, 2},
	tileset = {
		FURNACE_1_1 = {1,1},
		FURNACE_2_1 = {2,1},
		FURNACE_1_2 = {1,2},
		FURNACE_2_2 = {2,2},
	},
	leftuse = multiTilePlace
})

baseitem:new("DOOR", {
	texture = "door.png",
	color = {0.6, 0.4, 0.1},
	stack = 99,
	usedistance = 8,
	speed = 1/10,
	tilesize = {1, 3},
	tileset = {
		DOOR_1_1 = {1,1},
		DOOR_1_2 = {1,2},
		DOOR_1_3 = {1,3},
	},
	leftuse = multiTilePlace
})

baseitem:new("WORKBENCH", {
	texture = "workbench.png",
	color = {0.8, 0.6, 0.3},
	stack = 99,
	usedistance = 8,
	speed = 1/10,
	tilesize = {2, 1},
	tileset = {
		WORKBENCH_1_1 = {1,1},
		WORKBENCH_2_1 = {2,1},
	},
	leftuse = multiTilePlace,
})


baseitem:new("IRON_INGOT", {
	displayname = "IRON INGOT",
	texture = "ingot.png",
	color = {1, 0.8, 0.8}
})

baseitem:new("COPPER_INGOT", {
	displayname = "COPPER INGOT",
	texture = "ingot.png",
	color = {1, 0.45, 0.0}
})
baseitem:new("LEAD_INGOT", {
	displayname = "LEAD INGOT",
	texture = "ingot.png",
	color = {0.35, 0.35, 0.45}
})

baseitem:new("COAL", {

})

baseitem:new("PAPER", {

})

baseitem:new("BOTTLE", {
	stack = 99,
	texture = "potion.png",
	color = {1,1,1},
})

baseitem:new("ETHER", {
	tooltip = "Has a funny smell...",
	texture = "potion.png",
	color = {0.5, 0.5, 0.8},
	rarity = 2,
})

baseitem:new("CRYING_LILY", {
	texture = "potion.png",
	tooltip = "A peculiar flower that drips nectar as if it were weeping.",
	rarity = 2,
	color = {1,1,1},
})

baseitem:new("DEBUG_TOOL", {
	displayname = "DEBUG TOOL",
	speed = 1/10,
	stack = 1,
	repeating = false,
	rarity = -1,
	usedistance = 50,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

	leftuse = function(self, player)
		local mx, my = input.getTransformedMouse()

		local tx, ty = grid.pixelToTileXY(mx, my)

		local light = player.world:getLight(tx, ty)

		light = light + 0.1

		player.world:setLight(tx, ty, light)
		player.world:setLightUpdate(tx, ty, true)

		--player.world:tileUpdate(tx, ty)
		--player.world:randomTick(tx, ty)

		print("SETLIGHT "..tx..", "..ty..": "..light)
	end,
	rightuse = function(self, player)
		local mx, my = input.getTransformedMouse()

		local tx, ty = grid.pixelToTileXY(mx, my)

		print("TILEID "..tx..", "..ty..": ", player.world:rawget(tx, ty, "tiles"))
		print("STATE "..tx..", "..ty..": ", player.world:rawget(tx, ty, "states"))
		print("LIGHT "..tx..", "..ty..": ", player.world:rawget(tx, ty, "light"))
		print("DAMAGE "..tx..", "..ty..": ", player.world:rawget(tx, ty, "damage"))
		print("TILEUPDATE "..tx..", "..ty..": ", player.world:rawget(tx, ty, "tileupdate"))
	end,
})

baseitem:new("FERTILIZER", {
	displayname = "Fertilizer",
	speed = 1/10,
	repeating = true,
	specialdata = {usedistance=10},
	equip = function(self, player)
		player.showMouseTileDistance = 10
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = function(self, player)
		local mx, my = input.getTransformedMouse()

		local tx, ty = grid.pixelToTileXY(mx, my)

	end,
})



-- Add all tiles to the items list
local num = tiles:getNumberOfTiles()
for i = 1, num-1 do
	local data = tiles:getByID(i)
	baseitem:new(data.name.."_TILE", {
		displayname = string.gsub(data.name, "_", " "),
		texture = data.texture,
		color = data.color,
		speed = 1/6,
		animation = data.animation,
		animationspeed = data.animationspeed,
		tileid = data.id,
		usedistance = 8,
		inWorldScale = 1,
		playerHoldPosition = jutils.point.new(4, 4),
		swing = {
			start = -45,
			distance = 90,
		},
		defaultRotation = 0,
		stack = 999,
		equip = function(self, player)
			player.showMouseTileDistance = self.usedistance
		end,

		unequip = function(self, player)
			player.showMouseTileDistance = -1
		end,
		leftuse = function(self, player)
			local itemdata = player.itemEquipped
			local tiledata = tiles:getByID(self.tileid)


			local stack = player.itemEquippedStack
			local mx, my = input.getTransformedMouse()

			local tx, ty = grid.pixelToTileXY(mx, my)
			local px, py = getPlayerTile(player)
			local canPlace = true
			if magnitude(px, py, tx, ty) > self.usedistance then
				canPlace = false
			end

			if player.world:getTile(tx, ty) ~= 0 then
				canPlace = false
			end

			if tiledata.validplacement then
				if not tiledata.validplacement(player.world, tx, ty) then
					canPlace = false
				end
			end

			if canPlace then
				player.world:setTile(tx, ty, self.tileid)
				stack[2] = stack[2] - 1
				return true
			end
			return false
		end,
	})
end

items.ROPE_TILE.speed = 1/10
items.ROPE_TILE.leftuse = function(self, player)
	local itemdata = player.itemEquipped
	local tiledata = tiles:getByID(self.tileid)


	local stack = player.itemEquippedStack
	local mx, my = input.getTransformedMouse()

	local tx, ty = grid.pixelToTileXY(mx, my)
	local px, py = getPlayerTile(player)
	local canPlace = true
	if magnitude(px, py, tx, ty) > self.usedistance then
		canPlace = false
	end

	if player.world:getTile(tx, ty) ~= 0 then
		canPlace = false
	end

	if tiledata.validplacement then
		if not tiledata.validplacement(player.world, tx, ty) then
			canPlace = false
		end
	end

	if canPlace then
		player.world:setTile(tx, ty, self.tileid)
		stack[2] = stack[2] - 1
		return true
	end

	if player.world:getTile(tx, ty) == tiles.ROPE.id then
		local keepGoing = true
		local valid = true
		local iter = 1
		while keepGoing do
			if player.world:getTile(tx, ty+iter) == tiles.ROPE.id then
				iter = iter + 1
			elseif player.world:getTile(tx, ty+iter) == tiles.AIR.id then
				keepGoing = false
			else
				valid = false
				keepGoing = false
			end
		end

		if valid then
			player.world:setTile(tx, ty+iter, tiles.ROPE.id)
			stack[2] = stack[2] - 1
			return true
		end
	end


	return false
end



return itemcollector