local jutils = require("utility.jutils")
local config = require("config")
local object = require("utility.object")
local grid = require("utility.grid")

local tilecollector = {}
local tiles = {}
local idinc = 0

local defaultQuad = love.graphics.newQuad(0, 0, 8, 8, 8, 8)

do
	function tilecollector:tileHasTag(tileid, looktag)

		local tiledata

		if type(tileid) == "number" then
			tiledata = tilecollector:getByID(tileid)
		else
			tiledata = tileid
		end


		for _, tag in pairs(tiledata.tags) do
			if tag == looktag then
				return true
			end
		end
		return false
	end

	function tilecollector:getByID(tileid)
		assert(type(tileid=="number"), "TileID must be a number!")
		for name, data in pairs(tiles) do
			if data.id == tileid then
				return data
			end
		end
		error("No tile with ID".. tileid.. "exists!", 2)
	end

	function tilecollector:getNumberOfTiles()
		return idinc
	end

	local mt = {
		__index = function(t, k)

			if tiles[k] then return tiles[k] end
		end
	}

	setmetatable(tilecollector, mt)
end
----------------------------------

love.graphics.setDefaultFilter("nearest", "nearest")
---------------------------------------------------
-- tile update and random tile update callbacks

local function isSolid(tileid)
	local data = tilecollector:getByID(tileid)
	return data.solid
end

local function adjacentToNonSolidTile(world, x, y)
	for dx = -1, 1 do
		if not isSolid(world:getTile(x+dx, y)) then return true end	
	end
	for dy = -1, 1 do
		if not isSolid(world:getTile(x, y+dy)) then return true end
	end
	return false
end

local function grassCheck(world, x, y)

	local stayalive = adjacentToNonSolidTile(world, x, y)

	if not stayalive then world:setTile(x, y, tiles.DIRT.id) return end

	if world:getTile(x, y-1) == tiles.AIR.id then
		local poss = math.random()
		if poss > 0.95 then	
			world:setTile(x, y-1, tiles.CRYING_LILY.id)
		elseif poss > 0.5 then
			world:setTile(x, y-1, tiles.OVERGROWTH.id)
		end
	end

	if world:getTile(x, y+1) == tiles.AIR.id then
		local poss = math.random()
		if poss > 0.90 then	
			world:setTile(x, y+1, tiles.VINE.id)
		end
	end

	-- spreading onto dirt
	for dx = -1, 1 do
		for dy = -1, 1 do
			if world:getTile(x+dx, y+dy) == tiles.DIRT.id and adjacentToNonSolidTile(world, x+dx, y+dy) then
				world:setTile(x+dx, y+dy, tiles.GRASS.id)
			end
		end
	end
end

local function logValidCheck(world, x, y)
	if not isSolid(world:getTile(x, y+1)) then
		world:setTile(x, y, tiles.AIR.id, true)
	end
end

local function leavesRandomUpdate(world, x, y)
	local die = true

	for dx = -2, 2 do
		for dy = -2, 2 do
			if world:getTile(x+dx, y+dy) == tiles.LOG.id then
				die = false
			end
		end
	end

	if die then
		world:setTile(x, y, tiles.AIR.id)
	end
end

local function bluegrassCheck(world, x, y)
	local stayalive = adjacentToNonSolidTile(world, x, y)

	if not stayalive then world:setTile(x, y, tiles.MUD.id) return end

	-- spreading onto dirt
	for dx = -1, 1 do
		for dy = -1, 1 do
			if world:getTile(x+dx, y+dy) == tiles.MUD.id and adjacentToNonSolidTile(world, x+dx, y+dy) then
				world:setTile(x+dx, y+dy, tiles.BLUE_GRASS.id)
			end
		end
	end
end

local function explodeTNT(world, x, y)

	local state = world:getTileState(x, y)

	if state == 1 then
	
		local tntradius = 8
		local exp = world:addEntity("explosion", tntradius, tntradius+2, true)
		
		exp.position = jutils.point.new(x*config.TILE_SIZE, y*config.TILE_SIZE)
		for dx = -tntradius, tntradius do
			for dy = -tntradius, tntradius do
				if world:getTile(x+dx, y+dy) == tiles.TNT.id then
					world:setTileState(x+dx, y+dy, 1)
				end
			end
		end
	end
end

local function activationcheck(world, x, y)
	for dx = -1, 1 do
		if world:getTile(x+dx, y) == tiles.TNT.id then
			world:setTileState(x+dx, y, 1)
			world:setTile(x, y, tiles.AIR.id, false)
			--explodeTNT(world, x+dx, y)
			return
		end
	end
	for dy = -1, 1 do
		if world:getTile(x, y+dy) == tiles.TNT.id then
			world:setTileState(x, y+dy, 1)
			world:setTile(x, y, tiles.AIR.id, false)
			--explodeTNT(world, x, y+dy)
			return
		end
	end
	
end

---------------------------------------------------

-- Tile constructors
local function tile(name, data)

	local default = {
		name = name,
		id = idinc,
		color = {1, 1, 1},
		drop = name.."_TILE",
		texture = love.graphics.newImage("assets/default.png"),
		tags = {},
		quad = nil,
		collide = true,
		solid = true,
		animation = nil,
		animationspeed = 4, -- animation frames/sec
		tileupdate = nil,
		randomupdate = nil,
		validplacement = nil,
		onplace = nil,
		onbreak = nil,
		customrender = nil,
		hardness = 2,
		light = 0,
		absorb = 0.15,
	}

	local latestID = idinc
	if data.id~=nil then latestID = data.id end

	tiles[name] = {}
	for label, value in pairs(default) do
		tiles[name][label] = value
	end
	-- override default values
	for label, value in pairs(data) do
		if label == "texture" then
			tiles[name]["texture"] = love.graphics.newImage("assets/tiles/"..value)
		else
			tiles[name][label] = value
		end
	end

	idinc = latestID + 1
end

----------------------------------------------------
local function stoneTile(variantname, color)
	tile(variantname, {
		color = color,
		texture = "stone.png",
		hardness = 3,
	})
end

local function brickTile(variantname, color)
	tile(variantname, {
		color = color,
		texture = "brick.png",
		hardness = 4,
	})
end

-- Actual tiles begin here
-----------------------------------------------------------------------------

tile("VOID", {color = {0, 0, 0, 1}, texture = "blank.png", id = -1, solid = false, collide = false})
tile("AIR", {color = {0, 0, 0, 1}, texture = "blank.png", id = 0, light = 1.05, absorb = 0, solid = false, collide = false})


local function multitile(name, x, y, data)
	for tx = 1, x do
		for ty = 1, y do
			local dataTable = data
			dataTable.quad = love.graphics.newQuad((tx-1)*config.TILE_SIZE, (ty-1)*config.TILE_SIZE, 8, 8, (x)*config.TILE_SIZE, (y)*config.TILE_SIZE)
			dataTable.multitile = {
					name = name,
					ax = tx,
					ay = ty,
					w = x,
					h = y,
			}
			dataTable.tileupdate = function(world, x, y)

				local self = world:getTile(x, y)
				local tiledata = tilecollector:getByID(self)

				if data.validplacement then
					if data.validplacement(world, x, y) == false then
						world:setTile(x, y, tiles.AIR.id, false)
						return
					end
				end
					
				local multitile = tiledata.multitile

				local offsetx, offsety = multitile.w-(multitile.ax), multitile.h-(multitile.ay)

				for dx = 1, multitile.w do
					for dy = 1, multitile.h do
						local realx, realy = (x-multitile.ax)+dx, (y-multitile.ay)+dy

						local tile = world:getTile(realx, realy)
						local data = tilecollector:getByID(tile)

						if data.name ~= multitile.name.."_"..dx.."_"..dy then
							world:setTile(x, y, tiles.AIR.id, false)
							return
							end
						end
					end

				end
			
			tile(name.."_"..tx.."_"..ty, dataTable)
		end
	end
end

multitile("WORKBENCH", 2, 1, {
	texture = "workbench.png",
	color = {0.8, 0.6, 0.3},
	hardness = 3,
	solid = false,
	collide = true,
	tags = {"platformtile", "workbench"},
	drop = "WORKBENCH",
})

multitile("FURNACE", 2, 2, {
	texture = "furnace.png",
	color = {1,1,1},
	hardness = 4,
	light = 1.5,
	solid = false,
	collide = false,
	tags = {"furnace"},
	drop = "FURNACE"
})

multitile("CHEST", 2, 2, {

})

multitile("CAMPFIRE", 2, 1, {
	texture = "campfire.png",
	color = {1,1,1},
	light = 1.3,
	hardness = 3,
	solid = false,
	collide = false,
	drop = "CAMPFIRE",
})

tiles.CAMPFIRE_1_1.animation =  {
	[1] = love.graphics.newQuad(0, 0, 8, 8, 16, 16),
	[2] = love.graphics.newQuad(0, 8, 8, 8, 16, 16),
}

tiles.CAMPFIRE_2_1.animation = {
	[1] = love.graphics.newQuad(8, 0, 8, 8, 16, 16),
	[2] = love.graphics.newQuad(8, 8, 8, 8, 16, 16),
}

multitile("DOOR", 1, 3, {
	texture = "door.png",
	color = {0.8, 0.55, 0.3},
	hardness = 6,
	solid = true,
	collide = true,
	drop = "DOOR",
	-- fix how collisionboxes are read?
	collisionBox = {0, 0, 4, 0},
	playerInteract = function(player, x, y, button)
		if button == 2 then
			
			local self = player.world:getTile(x, y)

			local tiledata = tilecollector:getByID(self)

			local ypos = tiledata.multitile.ay
			local playerDir = player.direction

			for dy = 1, 3 do
				local boob =  y + ((dy) - ypos)
				if player.world:getTile(x+playerDir, boob) ~= 0 then return end
			end

			for dy = 1, 3 do
				local boob =  y + ((dy) - ypos)
				local tileat = player.world:getTile(x, boob)
				local tiledata = tilecollector:getByID(tileat)

				local close = 2
				local far = 1

				if playerDir == 1 then
					close = 1
					far = 2
				end

				local id = tiles["OPEN_DOOR_"..close.."_"..tiledata.multitile.ay].id
				player.world:setTile(x, boob, id, false)
				player.world:setTileState(x, boob, 1)

				local id2 = tiles["OPEN_DOOR_"..far.."_"..tiledata.multitile.ay].id
				player.world:setTile(x+playerDir, boob, id2, false)
			end

		end
	end
})

tiles.DOOR_1_3.validplacement = function(world, x, y)
	local below = world:getTile(x, y+1)

	if below < 1 then return false end

	local tiledata = tilecollector:getByID(below)

	if tiledata.solid == false then
		return false
	end
	return true
end

multitile("OPEN_DOOR", 2, 3, {
	texture = "opendoor.png",
	color = {0.8, 0.55, 0.3},
	hardness = 6,
	solid = false,
	collide = false,
	drop = "DOOR",
	playerInteract = function(player, x, y, button)
		if button == 2 then
			
			local playerx, playery = grid.pixelToTileXY(player.position.x, player.position.y)

			if playerx == x and playery == y then return end

			-- figure out which part of the door we are
			local self = player.world:getTile(x, y)

			local tiledata = tilecollector:getByID(self)

			local xpos = tiledata.multitile.ax
			local ypos = tiledata.multitile.ay
			
			for dx = 1, 2 do
				for dy = 1, 3 do
					local diffx = x + (dx-xpos)
					local diffy = y + (dy-ypos)
					
					local tileat = player.world:getTile(diffx, diffy)
					local tiledata = tilecollector:getByID(tileat)

					if player.world:getTileState(diffx, diffy) == 1 then
						player.world:setTile(diffx, diffy, tiles["DOOR_1_"..dy].id, false)
					else
						player.world:setTile(diffx, diffy, tiles.AIR.id, false)
					end
				end
			end

		end

	end
})

tile("GRASS", {
	texture = "soil.png",
	color = {0.2, 0.8, 0.2},
	randomupdate = grassCheck,
	tags = {"plantable-on"},
	drop = "DIRT_TILE",
})

tile("OVERGROWTH", {
	texture = "overgrowth.png",
	color = {0.2, 0.8, 0.2},
	randomupdate = function(world, x, y) end,
	tags = {"plant"},
	drop = nil,
	hardness = 1,
	solid = false,
	collide = false,
	tileupdate = function(world, x, y)
		if world:getTile(x, y+1) ~= tiles.GRASS.id then
			world:setTile(x, y, tiles.AIR.id, false)
		end
	end
})

tile("CRYING_LILY", {
	texture = "cryinglily.png",
	color = {1,1,1},
	hardness = 1,
	randomupdate = function(world, x, y) end,
	tags = {"plant"},
	drop = "CRYING_LILY",
	solid = false,
	collide = false,
	tileupdate = function(world, x, y)
		if world:getTile(x, y+1) ~= tiles.GRASS.id then
			world:setTile(x, y, tiles.AIR.id, false)
		end
	end
})

tile("VINE", {
	texture = "vine.png",
	color = {0.2, 0.8, 0.2},
	hardness = 1,
	randomupdate = function(world, x, y)
		if world:getTile(x, y+1) == tiles.AIR.id then
			world:setTile(x, y+1, tiles.VINE.id)
		end
	end,
	tags = {"plant"},
	drop = nil,
	solid = false,
	collide = false,
	tileupdate = function(world, x, y)
		local above = world:getTile(x, y-1) 
		if above ~= tiles.VINE.id and above ~= tiles.GRASS.id then
			world:setTile(x, y, tiles.AIR.id, false)
		end
	end
})

tile("DIRT", {
	texture = "soil.png",
	color = {0.45, 0.25, 0.1},
	tags = {"plantable-on"},
	hardness = 1,
})

tile("BLUE_GRASS", {
	texture = "soil.png",
	color = {0.2, 0.2, 0.9},
	randomupdate = bluegrassCheck,
	tags = {"plantable-on"},
	drop = "DIRT_TILE",
})

tile("MUD", {
	texture = "soil.png",
	color = {0.35, 0.15, 0.05},
	tags = {"plantable-on"},
	hardness = 1,
})

tile("ACTIVATOR", {
	tileupdate = activationcheck
})

tile("TNT", {
	texture = "tnt.png",
	color = {1, 1, 1},
	hardness = 1,
	tileupdate = explodeTNT,
})

local function torchValid(world, x, y)
	if not isSolid(world:getTile(x, y+1)) then
		return false
	end
	return true
end

tile("TORCH", {
	texture = "torch.png",
	color = {0.9, 0.9, 0.2},
	solid = false,
	collide = false,
	animation = {
		[1] = love.graphics.newQuad(0,0,8,8,32,8),
		[2] = love.graphics.newQuad(8,0,8,8,32,8),
		[3] = love.graphics.newQuad(16,0,8,8,32,8),
		[4] = love.graphics.newQuad(24,0,8,8,32,8),
	},
	light = 2,
	tileupdate = function(world, x, y)
		if not torchValid(world, x, y) then
			world:setTile(x, y, tiles.AIR.id, true)
		end
	end,
	validplacement = torchValid,
	hardness = 1,
})

local function ropeValid(world, x, y)
	if not (isSolid(world:getTile(x, y-1)) or world:getTile(x, y-1)==tiles.ROPE.id) then
		return false
	end
	return true
end

tile("ROPE", {
	texture = "rope.png",
	solid = false,
	collide = true,
	tags = {"rope"},

	tileupdate = function(world, x, y)
		if not ropeValid(world, x, y) then
			world:setTile(x, y, tiles.AIR.id, true)
		end
	end,
	customCollision = function(entity, separation, normal)
		if entity:isA("Player") then

			entity.touchingrope = true

		end
	end,
	validplacement = ropeValid,
	hardness = 1,
})
stoneTile("STONE",         {  1,   1,    1})
stoneTile("STONE_RED",     {  1, 0.8,  0.8})
stoneTile("STONE_YELLOW",  {  1,   1, 0.85})
stoneTile("STONE_BROWN",   {0.9, 0.9, 0.75})

tile("OBSIDIAN", {
	texture = "stone.png",
	color = {0.2, 0.2, 0.3},
	hardness = 10,
})

--[[tile("GLASS", {
	texture = "blank.png",
	color = {0.9, 0.9, 0.9, 0.2},
	hardness = 2,

})]]

brickTile("GRAY_BRICK",   {0.6, 0.6, 0.6})
brickTile("YELLOW_BRICK", {0.9, 0.9, 0.5})
brickTile("RED_BRICK",    {0.9, 0.5, 0.5})
brickTile("DARK_BRICK",   {0.2, 0.2, 0.2})
brickTile("WHITE_BRICK",  {  1,   1,   1})

tile("LOG", {
	texture = "log.png",
	color = {0.6, 0.4, 0.1},
	solid = true,
	collide = false,
	tileupdate = logValidCheck,
	drop = "PLANK_TILE",
	hardness = 8,
	tags = {"LOG"}
})

tile("LEAVES", {
	texture = "leaves.png",
	color = {0.2, 0.8, 0.2},
	solid = false,
	collide = false,
	randomupdate = leavesRandomUpdate,
	hardness = 1,
})
-- IDEA: dead leaves

tile("PLANK", {
	texture = "planks.png",
	color = {0.7, 0.5, 0.2},
	hardness = 3,
})

tile("PLATFORM", {
	texture = "platform.png",
	color = {0.7, 0.5, 0.2},
	solid = false,
	collide = true,
	tags = {"platformtile"}
})


local function plantValidityCheck(world, x, y)
	if tilecollector:tileHasTag(world:getTile(x, y+1), "plantable-on") then
		return true
		
	end
	return false
end

local function saplinggrow(world, x, y)
	
	local maketree = require("world.structures.maketree")

	local state = world:getTileState(x, y)+1

	if state > 8 then
		maketree(world, x, y+1)
		
	else
		world:setTileState(x, y, state)
	end
end

tile("SAPLING", {
	texture = "sapling.png",
	color = {1,1,1},
	solid = false,
	collide = false,
	tileupdate = function(world, x, y)
		if not plantValidityCheck then
			world:setTile(x, y, tiles.AIR.id, true)
		end
	end,
	randomupdate = saplinggrow,
	validplacement = plantValidityCheck,
	tags = {"plant"},
	hardness = 1,
})


tile("IRON_ORE", {
	texture = "ore.png",
	color = {1, 0.8, 0.8}
})

tile("COPPER_ORE", {
	texture = "ore.png",
	color = {1, 0.45, 0}
})

tile("LEAD_ORE", {
	texture = "ore.png",
	color = {0.35, 0.35, 0.45}
})

tile("SILVER_ORE", {
	texture = "ore.png",
	color = {1,1,1}
})

tile("PALLADIUM_ORE", {
	texture = "ore.png",
	color = {0.9, 0.5, 0.9},
})

tile("CHROMIUM_ORE", {
	texture = "ore.png",
	color = {0.5, 1, 1},
})

tile("GOLD_ORE", {
	texture = "ore.png",
	color = {1, 1, 0.5}
})

tile("TIN_ORE", {
	texture = "ore.png",
	color = {0.6, 0.4, 0.4}
})

tile("ALUMINIUM_ORE", {
	texture = "ore.png",
	color = {0.9, 0.9, 0.9}
})

tile("NICKEL_ORE", {
	texture = "ore.png",
	color = {1, 0.5, 0.5},
})

tile("VANADIUM_ORE", {
	texture = "ore.png",
	color = {0.5, 1, 0.8},
})

tile("COBALT_ORE", {
	texture = "ore.png",
	color = {0.3, 0.3, 1}
})

tile("TITANIUM_ORE", {
	texture = "ore.png",
	color = {0.6, 0.6, 0.6},
})

tile("URANIUM_ORE", {
	texture = "ore.png",
	color = {0.6, 1, 0.6}
})


tile("SAND", {
	texture = "soil.png",
	color = {0.7, 0.7, 0.3},
	hardness = 1,
})

tile("SMALL_BRICK", {
	texture = "bricksmall.png",
	color = {0.7, 0.7, 0.7}
})

tile("SLANTED_BRICK", {
	texture = "brickslanted.png",
	color = {1, 1, 1}
})

local function getWaterLevel(world, x, y)
	local id = world:getTile(x, y)

	if id == tiles.AIR.id then
		return 0
	end

	if id == tiles.WATER.id then
		return world:getTileState(x, y)
	end

	return -1
end

local function setWaterLevel(world, x, y, state)
	local id = world:getTile(x, y)

	if id == tiles.WATER.id then
		world:setTileState(x, y, state)
	end

	if id == tiles.AIR.id then
		world:setTile(x, y, tiles.WATER.id, false)
		world:setTileState(x, y, state)
	end
end

local waterLevels = 8

tile("WATER",{
	color = {0, 0, 0.5, 0.5},
	texture = "blank.png",
	solid = false,
	collide = true,
	onplace = function(world, x, y)
		world:rawset(x, y, "states", waterLevels)
	end,
	customCollision = function(entity, separation, normal)
		entity.velocity.x = jutils.math.clamp(-30, entity.velocity.x, 30)
		entity.velocity.y = jutils.math.clamp(-150, entity.velocity.y, 40)
	end,
	customrender = function(state, dmg, light, tx, ty)
		love.graphics.setColor(0, 0, 0.75, 0.5)
		love.graphics.rectangle("fill",
			(tx) * config.TILE_SIZE,
			(ty) * config.TILE_SIZE + (config.TILE_SIZE) - (state),
			config.TILE_SIZE,
			(config.TILE_SIZE / 8)*state)
	end,
	tileupdate = function(world, x, y)
		
		local waterMax = 8

		local this = getWaterLevel(world, x, y)
		if this < 1 then
			world:setTile(x, y, tiles.AIR.id, false)
			return
		end
		
		-- try to flow downward
		local below = getWaterLevel(world, x, y+1)
		local function check(val)
			if val ~= -1 and val < waterMax then return true end
		end

		if check(below) and this > 0 then

			local most = waterMax-below

			local taken = math.min(most, this)
			this = this-taken
			below = below+taken
			setWaterLevel(world, x, y, this)
			setWaterLevel(world, x, y+1, below)
			return
		end

		-- try to flow outwards once you're on the ground?
		local left = getWaterLevel(world, x-1, y)
		if check(left) and this > 0 and this~=left then
			if this > left+1 then
				this = this-1
				left = left+1
			elseif this-left == 1 then
				local rand = math.random()
				if rand > 0.49 then
					this = this
					left = this
				else
					this = left
					left = left
				end
			end
			setWaterLevel(world, x, y, this)
			setWaterLevel(world, x-1, y, left)
		end

		local right = getWaterLevel(world, x+1, y)
		if check(right) and this > 0 and this ~= right then
			if this > right+1 then
				this = this-1
				right = right+1
			elseif this-right == 1 then
				local rand = math.random()
				if rand > 0.49 then
					this = this
					right = this
				else
					this = right
					right = right
				end
			end
			setWaterLevel(world, x, y, this)
			setWaterLevel(world, x+1, y, right)
		end
	end,
})

local function getLavaLevel(world, x, y)
	local id = world:getTile(x, y)

	if id == tiles.AIR.id then
		return 0
	end

	if id == tiles.LAVA.id then
		return world:getTileState(x, y)
	end

	return -1
end

local function setLavaLevel(world, x, y, state)
	local id = world:getTile(x, y)

	if id == tiles.LAVA.id then
		world:setTileState(x, y, state)
	end

	if id == tiles.AIR.id then
		world:setTile(x, y, tiles.LAVA.id, false)
		world:setTileState(x, y, state)
	end
end


tile("LAVA", {
	color = {0.9, 0, 0, 0.95},
	light = 1.5,
	texture = "blank.png",
	solid = false,
	collide = true,
	onplace = function(world, x, y)
		world:rawset(x, y, "states", waterLevels)
	end,
	customCollision = function(entity, separation, normal)
		entity.velocity.x = jutils.math.clamp(-10, entity.velocity.x, 10)
		entity.velocity.y = jutils.math.clamp(-25, entity.velocity.y, 10)
		--entity.falling = false
	end,
	customrender = function(state, dmg, light, tx, ty)
		love.graphics.setColor(0.9, 0, 0, 0.95)
		love.graphics.rectangle("fill",
			(tx) * config.TILE_SIZE,
			(ty) * config.TILE_SIZE + (config.TILE_SIZE) - (state),
			config.TILE_SIZE,
			(config.TILE_SIZE / 8)*state)
	end,
	tileupdate = function(world, x, y)
		local lavaMax = 8

		local this = getLavaLevel(world, x, y)
		if this < 1 then
			world:setTile(x, y, tiles.AIR.id, false)
			return
		end

	
		for dx = -1, 1 do
			if world:getTile(x+dx, y) == tiles.WATER.id then
				if this == 8 then
					world:setTile(x+dx, y, tiles.AIR.id, false)
					world:setTile(x, y, tiles.OBSIDIAN.id, false)
				else
					world:setTile(x+dx, y, tiles.AIR.id, false)
					world:setTile(x, y, tiles.STONE.id, false)
				end
				return
			end
		end
		for dy = -1, 1 do
			if world:getTile(x, y+dy) == tiles.WATER.id then
				if this == 8 then
					world:setTile(x, y+dy, tiles.AIR.id, false)
					world:setTile(x, y, tiles.OBSIDIAN.id, false)
				else
					world:setTile(x, y+dy, tiles.AIR.id, false)
					world:setTile(x, y, tiles.STONE.id, false)
				end
				return
			end
		end
		
		-- try to flow downward
		local below = getLavaLevel(world, x, y+1)
		local function check(val)
			if val ~= -1 and val < lavaMax then return true end
		end

		if check(below) and this > 0 then
			this = this-1
			below = below+1
			setLavaLevel(world, x, y, this)
			setLavaLevel(world, x, y+1, below)
			return
		end

		-- try to flow outwards once you're on the ground?
		local left = getLavaLevel(world, x-1, y)
		if check(left) and this > 0 and this~=left then
			if this > left+1 then
				this = this-1
				left = left+1
			elseif this-left == 1 then
				local rand = math.random()
				if rand > 0.5 then
					this = this
					left = this
				else
					this = left
					left = left
				end
			end
			setLavaLevel(world, x, y, this)
			setLavaLevel(world, x-1, y, left)
		end

		local right = getLavaLevel(world, x+1, y)
		if check(right) and this > 0 and this ~= right then
			if this > right+1 then
				this = this-1
				right = right+1
			elseif this-right == 1 then
				local rand = math.random()
				if rand > 0.5 then
					this = this
					right = this
				else
					this = right
					right = right
				end
			end
			setLavaLevel(world, x, y, this)
			setLavaLevel(world, x+1, y, right)
		end
	end
})


-- End tile construction
----------------------------------
return tilecollector