local config = require("config")
local tiles = require("tiles")
local collision = require("utility.collision")
local jutils = require("utility.jutils")
local rendering = require("utility.rendering")
local items = require("items")

local grid = require("utility.grid")

----------------------------
-- World Class starts here

local world = {}

world.__index = world

function world.__init()
	local self = setmetatable({}, world)

	return self
end

function world.new(worldname)
	local self = world.__init()

	self.entitylist = {
		player = require("entities.player"),
		itemstack = require("entities.itemstack"),
		explosion = require("entities.explosion"),
		bee = require("entities.bee")
	}

	self.worldname = worldname
	self.entities = {}
	
	self.chunks = {}
	self.seed = 0
	self.chunkdrawradius = config.CHUNK_DRAW_RADIUS
	self.generationradius = 1
	self.focuspoint = jutils.point.new(0, 0)
	self.terrainGenerator = require("world.generators.terraingenerator")
	self.structureGenerator = require("world.generators.structuregenerator")
	self.chunknum = 0
	self.firstpassnum = 0
	self.secondpassnum = 0
	self.loadednum = 0
	self.fullnum = 0
	self.dx = 0
	self.dy = 0
	self.ambientlight = 0

	return self
end

function world:addEntity(entityname, ...)
	local entity = self.entitylist[entityname]:new(...)
	entity.world = self
	table.insert(self.entities, entity)
	return entity
end
function world:getMousePoint()
	return self.dx, self.dy
end

function world:setFocusPoint(point)
	self.focuspoint = point
end

---------------------------------------
-- World's map functionality

-- getters
function world:rawget(tilex, tiley, field, unsafe)

	assert(type(tilex) == "number")
	assert(type(tiley) == "number")
	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)

	if chunk then

		if chunk.loaded == false then
			if not (unsafe) then return -1 end
		end

		local exists = chunk[field]

		if exists then
			return chunk[field][lx][ly]
		end
	end
	return -1
end

function world:getTile(tilex, tiley)
	return self:rawget(tilex, tiley, "tiles", true)
end

function world:getTileState(tilex, tiley)
	return self:rawget(tilex, tiley, "states")
end

function world:getTileDamage(tilex, tiley)
	return self:rawget(tilex, tiley, "damage")
end

function world:getLight(tilex, tiley)
	return self:rawget(tilex, tiley, "light")
end

function world:damageTile(tilex, tiley, additionaldamage)
	local tile = self:rawget(tilex, tiley, "tiles")
	local maxdamage = tiles:getByID(tile).hardness
	local damage = self:rawget(tilex, tiley, "damage")

	local newdamage = damage + additionaldamage

	if newdamage >= maxdamage then
		self:setTile(tilex, tiley, 0, true)
	else
		self:rawset(tilex, tiley, "damage", newdamage)
	end
end

-- setters
function world:rawset(tilex, tiley, field, value, unsafe)

	
	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)

	if chunk then
		
		if chunk.loaded == false then
			if not (unsafe) then
				return -1
			end
		end

		local exists = chunk[field]

		if exists then
			chunk[field][lx][ly] = value
		end
	end
	return -1
end

function world:setTileDamage(tilex, tiley, damage)
	self:rawset(tilex, tiley, "damage", damage)
end

function world:setTileState(tilex, tiley, state)
	self:rawset(tilex, tiley, "states", state)

	self:tileUpdate(tilex, tiley - 1)
	self:tileUpdate(tilex, tiley + 1)
	self:tileUpdate(tilex + 1, tiley)
	self:tileUpdate(tilex - 1, tiley)
	self:tileUpdate(tilex, tiley)

end

-- sets the tile and fills the tileupdate flags of adjacent tiles
function world:setTile(tilex, tiley, tile, drop, ignorecallback)
	assert(type(tilex) == "number")
	assert(type(tiley) == "number")
	assert(type(tile) == "number")

	local current = self:getTile(tilex, tiley)
	if current ~= tile then
		self:setTileState(tilex, tiley, 0)

		--self:setLight(tilex, tiley, 0)
		self:setTileDamage(tilex, tiley, 0)
			
		if drop then
			local tiledata = tiles:getByID(current)
			if tiledata.drop ~= false then
				local item = items[tiledata.drop]

				local e = self:addEntity("itemstack", item.id, 1)
				e.nextposition.x = tilex*config.TILE_SIZE
				e.nextposition.y = tiley*config.TILE_SIZE
			end
		end

		self:rawset(tilex, tiley, "tiles", tile, true)

		local tiledata = tiles:getByID(tile)

		--chunk.light[lx][ly] = tiledata.light
		self:rawset(tilex, tiley, "lightupdate", true)
		self:setLightUpdate(tilex, tiley)
		if not ignorecallback then
			if tiledata.onplace then
				tiledata.onplace(self, tilex, tiley)
			end
		end
		

		self:tileUpdate(tilex, tiley)
		self:tileUpdate(tilex, tiley - 1)
		self:tileUpdate(tilex, tiley + 1)
		self:tileUpdate(tilex + 1, tiley)
		self:tileUpdate(tilex - 1, tiley)
			
	end
end

function world:setLight(tx, ty, level)
	self:rawset(tx, ty, "light", level)
end

function world:setLightUpdate(tx, ty)
	self:rawset(tx, ty, "lightupdate", true)
end

function world:applyLightRec(tx, ty, inputlight, recursions)
	recursions = recursions or 0
	if recursions > 500 then return end


	local minimumAbsorb = 0.05

	local tileid = self:rawget(tx, ty, "tiles")
	if tileid == -1 then return end

	

	local tiledata = tiles:getByID(tileid)
	local absorb = tiledata.absorb+minimumAbsorb
	local light = tiledata.light
	local current = self:getLight(tx, ty)

	local result = math.max(light, inputlight, current)

	if result == current then return end -- stop updating when the light levels match
	if result <= 0 then return end -- or when no light is left
	self:setLight(tx, ty, result)

	self:applyLightRec(tx+1, ty, result-absorb, recursions+1)
	self:applyLightRec(tx-1, ty, result-absorb, recursions+1)
	self:applyLightRec(tx, ty-1, result-absorb, recursions+1)
	self:applyLightRec(tx, ty+1, result-absorb, recursions+1)

end

--function world:lightUpdate(tx, ty)
	--self:applyLightRec(self, tx, ty, 0)
--end


-------------------------------------------------------------------------------
-- TILE AND CHUNK HANDLING

function world:getchunk(cx, cy)
	return self.chunks[grid.coordinatesToKey(cx, cy)]
end

function world:setchunkloaded(cx, cy)
	self.loaded[grid.coordinatesToKey(cx, cy)] = true
end

function world:tileUpdate(tilex, tiley)
	self:rawset(tilex, tiley, "tileupdate", true)
end

function world:randomTick(wx, wy)
	local tileid = self:getTile(wx, wy)
	local data = tiles:getByID(tileid)

	if data.randomupdate then
		data.randomupdate(self, wx, wy)
	end
end

function world:chunkRandomTileUpdates(chunk)
	local wx, wy = chunk.position.x, chunk.position.y
	for i = 1, config.RANDOM_TILE_UPDATES_PER_FRAME do
		local randx = math.random(1, config.CHUNK_SIZE)
		local randy = math.random(1, config.CHUNK_SIZE)
		self:randomTick((wx * config.CHUNK_SIZE) + randx, (wy * config.CHUNK_SIZE) + randy)
	end
end

function world:chunkTileUpdates(chunk)
	local count = 0
	local wx, wy = chunk.position.x, chunk.position.y
	for dx = 1, config.CHUNK_SIZE do
		for dy = 1, config.CHUNK_SIZE do
			-- the tile update
			if count > config.MAX_TILE_UPDATES_PER_FRAME then
				return
			end

			-- lighting update
			if chunk.lightupdate[dx][dy] == true then
				chunk.lightupdate[dx][dy] = false

				local light = tiles:getByID(chunk.tiles[dx][dy]).light

				self:applyLightRec((wx * config.CHUNK_SIZE) + dx - 1, (wy * config.CHUNK_SIZE) + dy - 1, light)
			end

			if chunk.tileupdate[dx][dy] == true then
				chunk.tileupdate[dx][dy] = false
				local tile = chunk.tiles[dx][dy]
				if tile > 0 then
					local data = tiles:getByID(tile)

					if data.tileupdate then
						data.tileupdate(self, (wx * config.CHUNK_SIZE) + dx - 1, (wy * config.CHUNK_SIZE) + dy - 1)

						if tile ~= tiles.WATER.id and tile ~= tiles.LAVA.id then
							count = count + 0.25
						else
							count = count + 1
						end
					end
				end
			end
		end
	end
end

local function newChunkStruct(cx, cy)
	return {
		position = jutils.point.new(cx, cy),
		tiles = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE), -- 2D array with default 0
		states = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE),
		damage = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE),
		tileupdate = jutils.table.construct(jutils.table.construct(false, config.CHUNK_SIZE), config.CHUNK_SIZE), -- 2D array with default false
		light = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE),
		lightupdate = jutils.table.construct(jutils.table.construct(false, config.CHUNK_SIZE), config.CHUNK_SIZE),
		requested = false, -- if the chunk has been requested by the world
		requestedFully = false,
		loaded = false, -- fully loaded
		terrainPass = false,
		structurePass = false,
		visible = false, -- to be rendered
	}
end

function world:map_clearChunkFields()
	for key, chunk in pairs(self.chunks) do
		chunk.requested = false
		chunk.visible = false
		chunk.requestedFully = false
	end
end

function world:map_fillChunkFields()
	local fx, fy = self.focuspoint.x, self.focuspoint.y
	
	local genradius = config.CHUNK_DRAW_RADIUS + config.CHUNK_SIMULATION_RADIUS + config.CHUNK_BUFFER_RADIUS

	for cx = -genradius, genradius do
		for cy = -genradius, genradius do
			local chunk = self:getchunk(fx+cx, fy+cy)

			if not chunk then
				local key = grid.coordinatesToKey(fx+cx, fy+cy)
				self.chunks[key] = newChunkStruct(fx+cx, fy+cy)
				chunk = self.chunks[key]
			end
			chunk.requested = true
		end
	end

	local simradius = config.CHUNK_DRAW_RADIUS + config.CHUNK_SIMULATION_RADIUS
	for cx = -simradius, simradius do
		for cy = -simradius, simradius do
			local chunk = self:getchunk(fx+cx, fy+cy)
			chunk.requestedFully = true
		end
	end

	local drawradius = config.CHUNK_DRAW_RADIUS
	for cx = -drawradius, drawradius do
		for cy = -drawradius, drawradius do
			self:getchunk(fx+cx, fy+cy).visible = true
		end
	end
end

function world:map_unloadOldChunks()
	-- cull chunks that are no longer "loaded"
	for key, chunk in pairs(self.chunks) do
		if chunk.requested == false then
			self.chunks[key] = nil
		end
	end
end

function world:map_loadNewChunks()
	for key, chunk in pairs(self.chunks) do
		-- run terrainPass on chunks with no terrain
		if chunk.requested == true and chunk.terrainPass == false then
			self.terrainGenerator(chunk)
			chunk.terrainPass = true
			return
		end

		if chunk.requestedFully == true and chunk.terrainPass == true and chunk.structurePass == false then
			local wx, wy = chunk.position.x, chunk.position.y
			for x = 1, config.CHUNK_SIZE do
				for y = 1, config.CHUNK_SIZE do
					self.structureGenerator(self, ((wx) * config.CHUNK_SIZE) + x - 1, ((wy) * config.CHUNK_SIZE) + y - 1)
				end
			end
			chunk.structurePass = true
			chunk.loaded = true
			self:applyLightRec(wx*config.CHUNK_SIZE + 1 , wy*config.CHUNK_SIZE + 1, 0)
			self:applyLightRec(wx*config.CHUNK_SIZE + 16 , wy*config.CHUNK_SIZE + 16, 0)
			return
		end
	end
end

function world:map_updateCurrentChunks()
	for key, chunk in pairs(self.chunks) do
		if chunk.visible == true and chunk.loaded == true then
			self:chunkTileUpdates(chunk)
			self:chunkRandomTileUpdates(chunk)
		end
	end
end

function world:drawChunks()
	for key, chunk in pairs(self.chunks) do
		if chunk.visible == true then

			local wx = chunk.position.x * config.CHUNK_SIZE
			local wy = chunk.position.y * config.CHUNK_SIZE

			for x = 1, config.CHUNK_SIZE do
				for y = 1, config.CHUNK_SIZE do
					local tileid = chunk.tiles[x][y]
					local damage = chunk.damage[x][y]
					local state = chunk.states[x][y]
					local light = chunk.light[x][y]

					if tileid > 0 then
						rendering:draw(tileid, state, damage, light, wx + x - 1, wy + y - 1)
					end
				end
			end
			if config.debug.DRAW_CHUNK_BORDERS then
				love.graphics.setColor(0.5, 0.5, 0.5)
				love.graphics.line((wx) * config.TILE_SIZE, (wy) * config.TILE_SIZE, ((wx) * config.TILE_SIZE) + (config.CHUNK_SIZE * config.TILE_SIZE), ((wy) * config.TILE_SIZE))
				love.graphics.line((wx) * config.TILE_SIZE, (wy) * config.TILE_SIZE, ((wx) * config.TILE_SIZE), ((wy) * config.TILE_SIZE) + config.CHUNK_SIZE * config.TILE_SIZE)
			end
		end
	end
end
----------------------------------------------------------------------------------
-- ENTITY AND COLLISION HANDLING


function world:updatePhysics(dt)

	-- ghetto weird hack to keep physics from going spastic
	if dt > (1/20) then dt = 1/20 end

	for idx, entity in pairs(self.entities) do
		if entity:isA("PhysicalEntity") then
			-- do collision solving
			entity:updatePhysics(dt)

			entity.falling = true
			
			local function testCollision(tilex, tiley)
				local tileid = self:getTile(tilex, tiley)
				if not (tileid > 0) then return false end
				
				local tiledata = tiles:getByID(tileid)
				if tiledata.collide == false then return false end

				local ex, ey, ew, eh = entity.nextposition.x, entity.nextposition.y, entity.boundingbox.x, entity.boundingbox.y

				local tx, ty, tw, th = (tilex*config.TILE_SIZE)+(config.TILE_SIZE/2), (tiley*config.TILE_SIZE)+(config.TILE_SIZE/2), config.TILE_SIZE/2, config.TILE_SIZE/2

				if tiledata.collisionBox then
					tx = tx + tiledata.collisionBox[1]
					ty = ty + tiledata.collisionBox[2]
					tw = tw - tiledata.collisionBox[3]
					th = th - tiledata.collisionBox[4]
				end

				local sx, sy = collision:test(ex, ey, ew, eh, tx, ty, tw, th)

				if not (sx and sy) then return false end

				local separation = jutils.point.new(sx, sy)

				local normalx, normaly = collision:solve(sx, sy, entity.velocity.x, entity.velocity.y)

				entity:collisionCallback(tileid, jutils.point.new(tilex, tiley), separation, jutils.point.new(normalx, normaly))
				if tiledata.customCollision then
					tiledata.customCollision(entity, separation, jutils.point.new(normalx, normaly), jutils.point.new(tilex, tiley))
				elseif tiledata.solid then-- default collision solver
					if normalx ~= nil and normaly ~= nil then
						entity.nextposition = entity.nextposition + separation
						-- collides with top of tile
						if normaly == -1 then
							entity.velocity.y = 0
							entity.falling = false
						end

						if normaly == 1 then
							entity.velocity.y = -(entity.velocity.y * 0.3)
						end
					end
					
				end
				
			end

			if entity.noclip == false then
				local entityTileX, entityTileY = grid.pixelToTileXY(entity.nextposition.x, entity.nextposition.y)
				local extx, exty = grid.pixelToTileXY(entity.boundingbox.x, entity.boundingbox.y)
				extx = extx + 1
				exty = exty + 1

				-- tile extents around the player
				for x = -extx, extx do
					for y = -exty, exty do
						testCollision(entityTileX+x, entityTileY+y)
					end
				end
			end



		end
	end
end

function world:updateEntities(dt)
	for idx, entity in pairs(self.entities) do
		entity:update(dt)

		-- remove entity references if they are dead (allow garbage collection)


		-- determine if entity is outside of loaded chunks (so we can kill it)
		local x, y = self.focuspoint.x*config.CHUNK_SIZE*config.TILE_SIZE, self.focuspoint.y*config.CHUNK_SIZE*config.TILE_SIZE
		local radius = config.CHUNK_DRAW_RADIUS+config.CHUNK_SIMULATION_RADIUS+config.CHUNK_BUFFER_RADIUS

		radius = (radius * config.CHUNK_SIZE) * config.TILE_SIZE

		if not entity:isA("Player") then
			if entity.position:distance(jutils.point.new(x, y)) > radius then
				entity.unloadtimer = entity.unloadtimer + dt
			else
				entity.unloadtimer = 0
			end

			if entity.unloadtimer > 1 then
				entity.dead = true
			end
		end

		if entity.dead then
			entity:dies()
			entity.world = nil
			self.entities[idx] = nil
			entity = nil
		end
	end
end

local ticker = 0

local tracker = 0

function world:update(dt)

	self:map_clearChunkFields()
	self:map_fillChunkFields()
	self:map_unloadOldChunks()
	self:map_loadNewChunks()

	self:map_updateCurrentChunks()

	self:updateEntities(dt)

	tracker = tracker + dt

	if tracker >= (1/80) then
		self:updatePhysics(tracker)
		tracker = 0
	end
end

function world:drawEntities()
	for _, entity in pairs(self.entities) do
		entity:draw()
	end
end

local SHOW_ENTITY_LIST = false

function world:draw()
	self.dx, self.dy = love.graphics.inverseTransformPoint(love.mouse.getX(), love.mouse.getY())

	self:drawChunks()
	self:drawEntities()
	if SHOW_ENTITY_LIST then
		love.graphics.push()
		love.graphics.origin()
		love.graphics.setColor(0,0,0)
		local inc = 0
		for idx, entity in pairs(self.entities) do
			love.graphics.print(entity.classname..", ".. tostring(entity), 800, inc*12)
			inc = inc + 1
		end
		love.graphics.pop()
	end
end

return world
